@@ -2,13 +2,11 @@
 available here as well to keep the number of imports to a minimum for most
 applications.
 """
-from __future__ import division
-from __future__ import absolute_import
-from __future__ import unicode_literals
-from __future__ import print_function
-__version__ = '3.0.0'
+from __future__ import absolute_import, division, print_function, unicode_literals
 
 from .base import SquirroClient
+from .document_uploader import DocumentUploader
 from .exceptions import *
 from .item_uploader import ItemUploader
-from .document_uploader import DocumentUploader
+
+__version__ = "3.2.0"
