@@ -1,9 +1,7 @@
 """Sends items to Squirro."""
-from __future__ import division
-from __future__ import absolute_import
-from __future__ import print_function
-
+from __future__ import absolute_import, division, print_function
 from future import standard_library
+
 standard_library.install_aliases()  # noqa
 
 import copy
@@ -11,6 +9,7 @@ import logging
 import math
 import os
 import sys
+import time
 from builtins import object, range
 from configparser import SafeConfigParser
 from numbers import Number
@@ -37,11 +36,13 @@ try:
 
     def get_client_cls():
         try:
-            return get_injected('squirro_client_cls')
+            return get_injected("squirro_client_cls")
         except DependencyNotFound:
             return SquirroClient
 
+
 except ImportError:
+
     def get_client_cls():
         return SquirroClient
 
@@ -79,6 +80,14 @@ class ItemUploader(object):
     :param pipeline_workflow_name: Pipeline workflow name. Either name or id
         need to be set.
     :param pipeline_workflow_id: Pipeline workflow ID.
+    :param non_retry_list: List of status codes for which we don't want a
+        retry/backoff logic.
+        Defaults to [200, 202, 401, 403, 400, 404]
+        200, 202: Successful codes.
+        401, 403: Already have a retry block in the _perform_request method
+                    of Squirro Client.
+        400, 404: Does not make sense to retry for these codes as retrying
+                    won't fix the underlying issue.
 
     Typical usage:
 
@@ -113,51 +122,72 @@ class ItemUploader(object):
 
     """
 
-    def __init__(self, token=None, project_id=None, project_title=None,
-                 object_id=None, source_name=None, source_ext_id=None,
-                 cluster=None, client_cls=None, batch_size=None,
-                 config_file=None, config_section=None,
-                 processing_config=None, steps_config=None,
-                 source_id=None, source_secret=None,
-                 pipeline_workflow_name=None, pipeline_workflow_id=None,
-                 timeout_secs=None, **kwargs):
-
+    def __init__(
+        self,
+        token=None,
+        project_id=None,
+        project_title=None,
+        object_id=None,
+        source_name=None,
+        source_ext_id=None,
+        cluster=None,
+        client_cls=None,
+        batch_size=None,
+        config_file=None,
+        config_section=None,
+        processing_config=None,
+        steps_config=None,
+        source_id=None,
+        source_secret=None,
+        pipeline_workflow_name=None,
+        pipeline_workflow_id=None,
+        timeout_secs=None,
+        non_retry_list=[200, 202, 400, 401, 403, 404],
+        **kwargs
+    ):
+        self.non_retry_list = non_retry_list
         if source_secret:
-            msg = '`source_secret` parameter is deperecated and is not ' \
-                  'necessary anymore.'
+            msg = (
+                "`source_secret` parameter is deperecated and is not "
+                "necessary anymore."
+            )
             deprecation(msg)
             raise NotImplementedError(msg)
 
         if object_id:
-            msg = '`object_id` parameter is deperecated and is not necessary ' \
-                  'anymore.'
+            msg = (
+                "`object_id` parameter is deperecated and is not necessary " "anymore."
+            )
             deprecation(msg)
             raise NotImplementedError(msg)
 
         if processing_config:
-            msg = '`processing_config` parameter is deperecated and is not ' \
-                  'supported anymore. Please use steps_config or ' \
-                  'pipeline_workflow_name or pipeline_workflow_id instead.'
+            msg = (
+                "`processing_config` parameter is deperecated and is not "
+                "supported anymore. Please use steps_config or "
+                "pipeline_workflow_name or pipeline_workflow_id instead."
+            )
             deprecation(msg)
             raise NotImplementedError(msg)
 
         # possibly deprecated bits
         self.config = None
         kwargs.update(locals())
-        del kwargs['self']
+        del kwargs["self"]
         self._load_config(kwargs)
         self._validate_config()
 
         # Do not create a new SqiurroClient if an authenticated client instance
         # has been supplied already
         self.client = None
-        if kwargs.get('client_instance'):
-            self.client = kwargs.get('client_instance')
+        if kwargs.get("client_instance"):
+            self.client = kwargs.get("client_instance")
         else:
             self._create_squirro_client(
                 client_cls,
                 retry_status_forcelist=frozenset([502, 503, 504]),
-                timeout_secs=timeout_secs)
+                timeout_secs=timeout_secs,
+            )
 
         self._lookup_project()
 
@@ -165,12 +195,13 @@ class ItemUploader(object):
             # Use existing source if exists
             # Otherwise fail
             # Warn ignoring fields if given
-            if any([pipeline_workflow_id, pipeline_workflow_name,
-                    steps_config]):
-                log.warn('Ignoring parameters: `pipeline_workflow_id`, '
-                         '`pipeline_workflow_name`, and `steps_config`')
+            if any([pipeline_workflow_id, pipeline_workflow_name, steps_config]):
+                log.warn(
+                    "Ignoring parameters: `pipeline_workflow_id`, "
+                    "`pipeline_workflow_name`, and `steps_config`"
+                )
 
-            self.source_id = self._lookup_source(source_id=source_id)['id']
+            self.source_id = self._lookup_source(source_id=source_id)["id"]
             return
 
         elif pipeline_workflow_id:
@@ -178,8 +209,10 @@ class ItemUploader(object):
             # Otherwise fail
             # Warn ignoring fields if given
             if any([pipeline_workflow_name, steps_config]):
-                log.warn('Ignoring parameters: `pipeline_workflow_name`, '
-                         'and `steps_config`')
+                log.warn(
+                    "Ignoring parameters: `pipeline_workflow_name`, "
+                    "and `steps_config`"
+                )
             workflow = self._lookup_workflow(uuid=pipeline_workflow_id)
 
         elif steps_config:
@@ -192,9 +225,10 @@ class ItemUploader(object):
                 workflow = self._lookup_workflow(name=pipeline_workflow_name)
 
             else:
-                pipeline_workflow_name = 'Item Uploader'
+                pipeline_workflow_name = "Item Uploader"
                 workflow = self.client.new_pipeline_workflow(
-                    self.project_id, pipeline_workflow_name, steps=steps_config)
+                    self.project_id, pipeline_workflow_name, steps=steps_config
+                )
 
         elif pipeline_workflow_name:
             # Use workflow with this name
@@ -208,23 +242,31 @@ class ItemUploader(object):
         # get or create source
         try:
             source = self._lookup_source(source_name=source_name)
-            if source['pipeline_workflow_id'] != workflow['id']:
+            if source["pipeline_workflow_id"] != workflow["id"]:
                 raise ValueError(
-                    'Specified source {} exists, but is incompatible with '
-                    'provided `pipeline_workflow_name`, `pipeline_workflow_id`'
-                    ', or `steps_config`'.format(source))
+                    "Specified source {} exists, but is incompatible with "
+                    "provided `pipeline_workflow_name`, `pipeline_workflow_id`"
+                    ", or `steps_config`".format(source)
+                )
         except NotFoundError:
-            log.info('Source %r in Project %r not present. Creating one',
-                     source_id, self.project_id)
-            source = self._create_source(source_name=source_name,
-                                         pipeline_workflow_id=workflow['id'])
-            log.info('New source created with id %r', source['id'])
+            log.info(
+                "Source %r in Project %r not present. Creating one",
+                source_id,
+                self.project_id,
+            )
+            source = self._create_source(
+                source_name=source_name, pipeline_workflow_id=workflow["id"]
+            )
+            log.info("New source created with id %r", source["id"])
 
-        self.source_id = source['id']
+        self.source_id = source["id"]
 
         log.info(
-            'Using source id %r with pipeline workflow id %r in project %r',
-            self.source_id, workflow['id'], self.project_id)
+            "Using source id %r with pipeline workflow id %r in project %r",
+            self.source_id,
+            workflow["id"],
+            self.project_id,
+        )
 
     def _lookup_workflow(self, name=None, uuid=None):
         """Returns the workflow for a given workflow `name` or `uuid`.
@@ -234,60 +276,57 @@ class ItemUploader(object):
         If no workflow is found a NotFoundError is raised."""
         workflows = self.client.get_pipeline_workflows(self.project_id)
 
-        if uuid and uuid != 'default':
+        if uuid and uuid != "default":
             for workflow in workflows:
-                if workflow['id'] == uuid:
+                if workflow["id"] == uuid:
                     return workflow
 
         elif name:
             for workflow in workflows:
-                if workflow['name'] == name:
+                if workflow["name"] == name:
                     return workflow
 
-        elif not uuid or uuid == 'default':
+        elif not uuid or uuid == "default":
             for workflow in workflows:
-                if workflow['project_default']:
+                if workflow["project_default"]:
                     return workflow
 
-        raise NotFoundError('Workflow "%s" with id "%s" was not found!' %
-                            (name, uuid))
+        raise NotFoundError('Workflow "%s" with id "%s" was not found!' % (name, uuid))
 
     def _load_config(self, kwargs):
         """Loads the configuration from file and merges it with the passed-in
         arguments"""
         parser = SafeConfigParser()
-        file_names = kwargs.get('config_file')
+        file_names = kwargs.get("config_file")
         if not file_names:
-            file_names = [os.path.expanduser('~/.squirrorc')]
+            file_names = [os.path.expanduser("~/.squirrorc")]
         parser.read(file_names)
 
         # initialize with defaults
         config = copy.deepcopy(self.DEFAULTS)
 
         # read from config-file
-        config_section = kwargs.get('config_section')
+        config_section = kwargs.get("config_section")
         if not config_section:
-            config_section = 'squirro'
+            config_section = "squirro"
         if parser.has_section(config_section):
             config.update(dict(parser.items(config_section)))
 
         # update with kwargs actually passed in
-        config.update(dict([(k, v) for k, v in kwargs.items()
-                            if v is not None]))
-        config = dict([(k, v) for k, v in config.items()
-                       if k in self.CONFIG_KEYS])
+        config.update(dict([(k, v) for k, v in kwargs.items() if v is not None]))
+        config = dict([(k, v) for k, v in config.items() if k in self.CONFIG_KEYS])
 
         # transform integer arguments
         for key in self.INT_OPTIONS:
-            if not config.get('key') is None:
+            if not config.get("key") is None:
                 config[key] = int(config[key])
 
         # adjust the cluster URL
-        if config.get('cluster'):
-            config['cluster'] = config['cluster'].rstrip('/')
+        if config.get("cluster"):
+            config["cluster"] = config["cluster"].rstrip("/")
 
-        if not config.get('source_ext_id'):
-            config['source_ext_id'] = config.get('source_name')
+        if not config.get("source_ext_id"):
+            config["source_ext_id"] = config.get("source_name")
 
         self.config = config
 
@@ -295,64 +334,72 @@ class ItemUploader(object):
         """Validates mandatory parameters and checks invalid combination.
         Raises a `ValueError` if something is off"""
 
-        if not any([self.config.get('project_id'),
-                    self.config.get('project_title')]):
+        if not any([self.config.get("project_id"), self.config.get("project_title")]):
+            raise ValueError('Parameters ("project_id"/"project_title") is required')
+
+        if self.config.get("source_id") and self.config.get("steps_config"):
             raise ValueError(
-                'Parameters ("project_id"/"project_title") is required')
+                'Parameters "source_id" and "steps_config" ' "are mutually exclusive."
+            )
 
-        if self.config.get('source_id') and self.config.get('steps_config'):
-            raise ValueError('Parameters "source_id" and "steps_config" '
-                             'are mutually exclusive.')
-
-        if not self.config.get('token'):
+        if not self.config.get("token"):
             raise ValueError('Mandatory parameter "token" is missing')
 
-        manual_config = [self.config.get('topic_api_url'),
-                         self.config.get('user_api_url'),
-                         self.config.get('provider_api_url')]
-        if not self.config.get('cluster') and not all(manual_config):
+        manual_config = [
+            self.config.get("topic_api_url"),
+            self.config.get("user_api_url"),
+            self.config.get("provider_api_url"),
+        ]
+        if not self.config.get("cluster") and not all(manual_config):
             raise ValueError('Mandatory parameter "cluster" is missing')
 
-        if all([self.config.get('project_id'),
-                self.config.get('project_title')]):
-            raise ValueError('Parameters "project_id" and "project_title" are '
-                             'mutually exclusive')
+        if all([self.config.get("project_id"), self.config.get("project_title")]):
+            raise ValueError(
+                'Parameters "project_id" and "project_title" are ' "mutually exclusive"
+            )
 
     def _create_squirro_client(self, client_cls, **kwargs):
         """Creates a Squirro client"""
-        if all([self.config.get('topic_api_url'),
-                self.config.get('user_api_url'),
-                self.config.get('provider_api_url')]):
-            kwargs.update({
-                'topic_api_url': self.config['topic_api_url'],
-                'user_api_url': self.config['user_api_url']
-            })
+        if all(
+            [
+                self.config.get("topic_api_url"),
+                self.config.get("user_api_url"),
+                self.config.get("provider_api_url"),
+            ]
+        ):
+            kwargs.update(
+                {
+                    "topic_api_url": self.config["topic_api_url"],
+                    "user_api_url": self.config["user_api_url"],
+                }
+            )
         else:
-            kwargs['cluster'] = self.config['cluster']
+            kwargs["cluster"] = self.config["cluster"]
 
         if not client_cls:
             client_cls = get_client_cls()
 
         self.client = client_cls(None, None, **kwargs)
-        self.client.authenticate(refresh_token=self.config['token'])
+        self.client.authenticate(refresh_token=self.config["token"])
 
     def _lookup_project(self):
         """Looks up the project either by `project_id` or `project_title`"""
-        if self.config.get('project_id'):
-            project_id = self.config['project_id']
+        if self.config.get("project_id"):
+            project_id = self.config["project_id"]
             self.client.get_project(project_id)
             self.project_id = project_id
         else:
-            project_title = self.config['project_title']
+            project_title = self.config["project_title"]
             projects = self.client.get_user_projects()
             for project in projects:
-                if project['title'] == project_title:
-                    self.project_id = project['id']
+                if project["title"] == project_title:
+                    self.project_id = project["id"]
                     break
 
-            if getattr(self, 'project_id', None) is None:
-                raise NotFoundError('No project with title {0!r} found'
-                                    .format(project_title))
+            if getattr(self, "project_id", None) is None:
+                raise NotFoundError(
+                    "No project with title {0!r} found".format(project_title)
+                )
 
     def _lookup_source(self, source_id=None, source_name=None):
         """Looks up a source with `source_id` called `source_name`."""
@@ -360,57 +407,76 @@ class ItemUploader(object):
         if source_id:
             try:
                 source = self.client.get_source(self.project_id, source_id)
-                if source_name and source.get('name') != source_name:
+                if source_name and source.get("name") != source_name:
                     raise ValueError(
-                        'Source %r in project %r does not have name %r',
-                        source_id, self.project_id, source_name)
+                        "Source %r in project %r does not have name %r",
+                        source_id,
+                        self.project_id,
+                        source_name,
+                    )
             except NotFoundError:
-                log.exception('Source %r in project %r not present.',
-                              source_id, self.project_id)
+                log.exception(
+                    "Source %r in project %r not present.", source_id, self.project_id
+                )
                 raise
 
         elif source_name:
-            sources = [source
-                       for source in self.client.get_sources(self.project_id)
-                       if source['name'] == source_name]
-            source_ids = [source['id'] for source in sources]
-            log.debug('Sources %r with name %r', source_ids, source_name)
+            sources = [
+                source
+                for source in self.client.get_sources(self.project_id)
+                if source["name"] == source_name
+            ]
+            source_ids = [source["id"] for source in sources]
+            log.debug("Sources %r with name %r", source_ids, source_name)
             if len(source_ids) > 1:
                 raise ValueError(
-                    'Multiple sources with ids {} found with name {} in '
-                    ' project {}. Please set the source_id parameter'.format(
-                        source_ids, source_name, self.project_id))
+                    "Multiple sources with ids {} found with name {} in "
+                    " project {}. Please set the source_id parameter".format(
+                        source_ids, source_name, self.project_id
+                    )
+                )
             elif len(source_ids) == 1:
                 source = sources[0]
             else:
                 raise NotFoundError(
-                    'Source with name {} in project {} not found.'.format(
-                        source_name, self.project_id))
+                    "Source with name {} in project {} not found.".format(
+                        source_name, self.project_id
+                    )
+                )
 
         else:
-            raise NotFoundError(
-                'Must define either `source_id` or `source_name`.')
+            raise NotFoundError("Must define either `source_id` or `source_name`.")
 
         return source
 
     def _create_source(self, source_name, pipeline_workflow_id):
         source_config = {
-            'dataloader_options': {
-                'project_id': self.project_id,
-                'plugin_name': 'ITEMUPLOADER',
-                'runs_on_client': True},
-            'dataloader_plugin_options': {}
+            "dataloader_options": {
+                "project_id": self.project_id,
+                "plugin_name": "ITEMUPLOADER",
+                "runs_on_client": True,
+            },
+            "dataloader_plugin_options": {},
         }
 
         res = self.client.new_source(
             project_id=self.project_id,
-            name=source_name or self.DEFAULTS['source_name'],
+            name=source_name or self.DEFAULTS["source_name"],
             config=source_config,
-            scheduling_options={'force_stop_scheduling': True},
-            pipeline_workflow_id=pipeline_workflow_id)
+            scheduling_options={"force_stop_scheduling": True},
+            pipeline_workflow_id=pipeline_workflow_id,
+        )
         return res
 
-    def upload(self, items, priority=0, pipeline_workflow_id=None):
+    def upload(
+        self,
+        items,
+        priority=0,
+        pipeline_workflow_id=None,
+        num_retries=10,
+        delay=1,
+        backoff=2,
+    ):
         """Sends ``items`` to Squirro.
 
         :param items: A list of items. See
@@ -422,42 +488,65 @@ class ItemUploader(object):
         :param pipeline_workflow_id: str, id of an existing pipeline
             workflow which should be used to process the current batch of
             items. Can only be used with parameter `priority` set to 1.
+        :param num_retries: int, Number of retries to make when a service is
+            unavailable.
+        :param delay: int, Initial delay in seconds between retries.
+        :param backoff: int, Backoff multiplier, e.g. value of 2 will double
+            the delay each retry.
         """
 
         self._validate_upload_parameters(priority, pipeline_workflow_id)
-        if not self.config.get('provider_api_url'):
-            cluster = self.config['cluster']
-            url = '{0}/api/provider/v1/{1}/projects/{2}/sources/{3}/push_items'.format(
-                cluster, self.client.tenant, self.project_id, self.source_id)
+        if not self.config.get("provider_api_url"):
+            cluster = self.config["cluster"]
+            url = "{0}/api/provider/v1/{1}/projects/{2}/sources/{3}/push_items".format(
+                cluster, self.client.tenant, self.project_id, self.source_id
+            )
         else:
-            endpoint = self.config['provider_api_url']
-            url = '{0}/v1/{1}/projects/{2}/sources/{3}/push_items'.format(
-                endpoint, self.client.tenant, self.project_id, self.source_id)
-
-        headers = {'Content-Type': 'application/json'}
-        params = {'priority': priority}
+            endpoint = self.config["provider_api_url"]
+            url = "{0}/v1/{1}/projects/{2}/sources/{3}/push_items".format(
+                endpoint, self.client.tenant, self.project_id, self.source_id
+            )
+        headers = {"Content-Type": "application/json"}
+        params = {"priority": priority}
         if pipeline_workflow_id is not None:
-            params.update({'pipeline_workflow_id': pipeline_workflow_id})
+            params.update({"pipeline_workflow_id": pipeline_workflow_id})
 
         batch_size = self._get_batch_size(items)
 
         if not items:
-            log.info('Asked to upload empty item list. Skipping.')
-
+            log.info("Asked to upload empty item list. Skipping.")
+        num_batches = math.ceil(len(items) / batch_size)
         for idx in range(0, len(items), batch_size):
-            for item in items[idx:idx + batch_size]:
+            mdelay = delay
+            for item in items[idx : idx + batch_size]:
                 self._fixup_keywords(item)
-            data = json.dumps(items[idx:idx + batch_size])
-            res = self.client._perform_request('post', url, params=params,
-                                               data=data, headers=headers)
-            self.client._process_response(res, [202])
+            data = json.dumps(items[idx : idx + batch_size])
+            for i in range(num_retries + 1):
+                if i != 0:
+                    log.warning("Retrying {} of {}".format(i, num_retries))
+                res = self.client._perform_request(
+                    "post", url, params=params, data=data, headers=headers
+                )
+                if res.status_code in self.non_retry_list:
+                    self.client._process_response(res, [202])
+                    break
+                else:
+                    log.warning(
+                        "Failed to upload batch {} of {}, returned "
+                        "status code {}. Sleeping for {} seconds.".format(
+                            idx + 1, num_batches, res.status_code, mdelay
+                        )
+                    )
+                    time.sleep(mdelay)
+                    mdelay *= backoff
 
     def _validate_upload_parameters(self, priority, pipeline_workflow_id):
         """Validations for parameters to the `upload` function."""
         if priority != 1 and pipeline_workflow_id:
             raise ValueError(
-                'Can not specify a `pipeline_workflow_id` in priority'
-                ' {}'.format(priority))
+                "Can not specify a `pipeline_workflow_id` in priority"
+                " {}".format(priority)
+            )
 
         if pipeline_workflow_id:
             self._lookup_workflow(uuid=pipeline_workflow_id)
@@ -466,10 +555,10 @@ class ItemUploader(object):
         """Gets the batch size from config or calculates the optimal size based
         on the items if batch_size is set to -1.
         """
-        MAX_BYTE_SIZE = 5e+7  # 50MB, allows for some json dumps overhead and
+        MAX_BYTE_SIZE = 5e7  # 50MB, allows for some json dumps overhead and
         # ES bulk action instructions
 
-        batch_size = self.config['batch_size']
+        batch_size = self.config["batch_size"]
 
         if batch_size == -1:
             items_bytes_size = sys.getsizeof(json.dumps(items))
@@ -479,8 +568,7 @@ class ItemUploader(object):
                 # how many chunks do we split the items into?
                 nof_chunks = math.ceil(float(items_bytes_size) / MAX_BYTE_SIZE)
                 assert nof_chunks >= 1
-                batch_size = \
-                    max(1, int(math.floor(float(len(items)) / nof_chunks)))
+                batch_size = max(1, int(math.floor(float(len(items)) / nof_chunks)))
 
         # Never return a batch_size of zero.
         if batch_size < 1:
@@ -490,10 +578,10 @@ class ItemUploader(object):
 
     def _fixup_keywords(self, item):
         """Validates and fixes keywords of `item`"""
-        if item.get('keywords') is None:
+        if item.get("keywords") is None:
             return
 
-        keywords = item['keywords']
+        keywords = item["keywords"]
         for key, value in keywords.items():
             if isinstance(value, (Number, six.string_types)):
                 keywords[key] = [value]
@@ -502,25 +590,32 @@ class ItemUploader(object):
             elif isinstance(value, set):
                 keywords[key] = list(value)
             else:
-                raise TypeError('Cannot validate the keyword {0}. Type {1} '
-                                'is not supported'.format(key, type(value)))
+                raise TypeError(
+                    "Cannot validate the keyword {0}. Type {1} "
+                    "is not supported".format(key, type(value))
+                )
 
     # default values for arguments
     DEFAULTS = {
-        'source_name': 'Upload',
-        'cluster': 'https://demo.squirro.net',
-        'batch_size': -1,  # auto calculate
+        "source_name": "Upload",
+        "cluster": "https://demo.squirro.net",
+        "batch_size": -1,  # auto calculate
     }
 
     # options to be converted to int
-    INT_OPTIONS = [
-        'batch_size',
-    ]
+    INT_OPTIONS = ["batch_size"]
 
     # arguments that can be passed in the constructor/ini-file
     CONFIG_KEYS = [
-        'token', 'cluster', 'project_id', 'project_title',
-        'source_name', 'source_ext_id', 'batch_size',
-        'user_api_url', 'topic_api_url', 'provider_api_url',
-        'source_id'
+        "token",
+        "cluster",
+        "project_id",
+        "project_title",
+        "source_name",
+        "source_ext_id",
+        "batch_size",
+        "user_api_url",
+        "topic_api_url",
+        "provider_api_url",
+        "source_id",
     ]
